# '==================================================================================================================================================================
# 'Disclaimer
# 'The sample scripts are not supported under any N-able support program or service. 
# 'The sample scripts are provided AS IS without warranty of any kind. 
# 'N-able further disclaims all implied warranties including, without limitation, any implied warranties of merchantability or of fitness for a particular purpose. 
# 'The entire risk arising out of the use or performance of the sample scripts and documentation stays with you. 
# 'In no event shall N-able or anyone else involved in the creation, production, or delivery of the scripts be liable for any damages whatsoever 
# '(including, without limitation, damages for loss of business profits, business interruption, loss of business information, or other pecuniary loss) 
# 'arising out of the use of or inability to use the sample scripts or documentation.
# '==================================================================================================================================================================

Param (
    [string]$verbose = "Y",
    [string]$applianceType = ""
)

function setupLogging() {
    $script:logFilePath = "C:\ProgramData\MspPlatform\Tech Tribes\reinstall AM Agent\debug.log"
	
	$script:logFolder = Split-Path $logFilePath
	$script:logFile = Split-Path $logFilePath -Leaf

    $logFolderExists = Test-Path $logFolder
    $logFileExists = Test-Path $logFilePath

    If ($logFolderExists -eq $false) {
        New-Item -ItemType "directory" -Path $logFolder | Out-Null
    }
    
    If ($logFileExists -eq $true) {
		Remove-Item $logFilePath -ErrorAction SilentlyContinue
		Start-Sleep 2
		New-Item -ItemType "file" -Path $logFolder -Name $logFile | Out-Null
    } Else {
		New-Item -ItemType "file" -Path $logFolder -Name $logFile | Out-Null
    }
    
    If (($logFolder -match '.+?\\$') -eq $false) {
        $script:logFolder = $logFolder + "\"
    }
}

function validateUserInput() {
    If ($verbose.ToLower() -eq "y") {
        $script:verboseMode = $true
        writeToLog V "You have defined to have the script output the verbose log entries."
    } Else {
        $script:verboseMode = $false
        writeToLog I "Will output logs in regular mode."
    }

    If ($applianceType.ToLower() -eq "agent") {
        writeToLog V "You have defined that the target device is a `"Windows Agent`"."
        $global:applianceType = "Agent"
    } ElseIf ($applianceType.ToLower() -eq "probe") {
        writeToLog V "You have defined that the target device is a `"Windows Software Probe`"."
        $global:applianceType = "Probe"
    } Else {
        writeToLog F "Appliance type was not defined. Please ensure you specify if this is for an `"agent`" or `"probe`"."
        writeToLog F "Failing script."
        Exit 1001
    }

    writeToLog I "Detected appliancetype: $applianceType"
    writeToLog V ("### Completed running {0} function. ###`r`n" -f $MyInvocation.MyCommand)
}

function getAgentPath($applianceType) {
	writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    If ($applianceType -eq "Agent") {
        $script:applianceName = "Windows Agent"
    } ElseIf ($applianceType -eq "Probe") {
        $script:applianceName = "Windows Software Probe"
    } Else {
        writeToLog F "Invalid appliance type."
        Exit 1001
    }

    try {
        $Keys = Get-ChildItem HKLM:\SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog W "Error occurred during the lookup of the CurrentVersion\Uninstall Path in the registry, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        writeToLog W "Will continue with validating agent path."
    }
	
	$Items = $Keys | Foreach-Object {
        Get-ItemProperty $_.PsPath
    }

    $script:amRegLocation = "HKLM:\SOFTWARE\N-Able Technologies\$applianceName"

	ForEach ($Item in $Items) {
        If ($Item.DisplayName -like $applianceName) {
            $script:localFolder = $Item.installLocation
            $script:registryPath = $Item.PsPath
            $registryName = $Item.PSChildName
			$script:uninstallString = $Item.UninstallString
            $registryDisplayName = $Item.DisplayName
            $registryVersion = $Item.DisplayVersion
            $registryInstallDate = $Item.InstallDate
            break
        }
    }

    If ($null -ne $localFolder) { 
        try {
            $script:amRegVersion = Get-ItemPropertyValue $amRegLocation -Name "AutomationManagerVersion" -ErrorAction Stop
        } catch {
            $msg = $_.Exception.Message
            $line = $_.InvocationInfo.ScriptLineNumber
            writeToLog W "Error occurred during the lookup of the AutomationManagerVersion Path in the registry, due to:`r`n`t$msg"
            writeToLog V "This occurred on line number: $line"
            $script:amRegKeyFound = $false
        }
        If ($amRegKeyFound -ne $false) {
            $script:amRegKeyFound = $true
            writeToLog I "Current version of AutomationManagerAgent: $amRegVersion"
        }
    }
    
    # Check WOW6432Node

    try {
        $Keys = Get-ChildItem HKLM:\SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog W "Error occurred during the lookup of the CurrentVersion\Uninstall Path in the registry, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        writeToLog W "Will continue with validating agent path."
    }

    $Items = $Keys | Foreach-Object {
        Get-ItemProperty $_.PsPath
    }

    $script:amRegLocation = "HKLM:\SOFTWARE\WOW6432Node\N-Able Technologies\$applianceName"

	ForEach ($Item in $Items) {
		If ($Item.DisplayName -like $applianceName) {
            $script:localFolder = $Item.installLocation
            $script:registryPath = $Item.PsPath
            $registryName = $Item.PSChildName
            $registryDisplayName = $Item.DisplayName
            $registryVersion = $Item.DisplayVersion
            $registryInstallDate = $Item.InstallDate
            break
        }
	}


    try {
        $script:amRegVersion = Get-ItemPropertyValue $amRegLocation -Name "AutomationManagerVersion" -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog W "Error occurred during the lookup of the AutomationManagerVersion Path in the registry, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        $script:amRegKeyFound = $false
    }
    If ($amRegKeyFound -ne $false) {
        $script:amRegKeyFound = $true
        writeToLog I "Current version of AutomationManagerAgent: $amRegVersion"
    }

    If (!$script:localFolder) {
        writeToLog F "No $applianceName located."
        writeToLog F "Failing script."
        Exit 1001
    } Else {
		$script:agentPresent = $true

		If (($script:localFolder -match '.+?\\$') -eq $false) {
			$script:localFolder = $script:localFolder + "\"
		}
	
		writeToLog I "Agent install location found:`r`n`t$localFolder"
        writeToLog V "Version:`r`n`t$registryVersion`r`nInstallation Date:`r`n`t$registryInstallDate"
		writeToLog V "Registry location: `r`n`t$registryPath`r`n`t$registryName"
	}

	writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function stopServices() {
	writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

	$array = @()
    If ($applianceType -eq "Agent") {
        $array += "Windows Agent Maintenance Service"
        $array += "Windows Agent Service"
    } ElseIf ($applianceType -eq "Probe") {
        $array += "Windows Software Probe Maintenance Service"
        $array += "Windows Software Probe Service"
    }
	$array += "AutomationManagerAgent"
	
	foreach ($serviceName in $array) {

		If (Get-Service $serviceName -ErrorAction SilentlyContinue) {
			writeToLog I "Detected the $serviceName service on the device, will now stop the service."

			try {
   				$script:stopService = Stop-Service -Name $serviceName -ErrorAction Stop -WarningAction SilentlyContinue
  			} catch {
				$msg = $_.Exception
				$line = $_.InvocationInfo.ScriptLineNumber
				writeToLog W "Failed to stop $serviceName, due to:`r`n`t$($msg.Message)"
				writeToLog V "This occurred on line number: $line"
				writeToLog V "Status:`r`n`t$($msg.Status)`r`nResponse:`r`n`t$($msg.Response)`r`nInner Exception:`r`n`t$($msg.InnerException)`r`n`r`nHResult: $($msg.HResult)`r`n`r`nTargetSite and StackTrace:`r`n$($msg.TargetSite)`r`n$($msg.StackTrace)`r`n"
			}

			writeToLog I "$serviceName service is now stopped."

		} Else {
			writeToLog W "$serviceName service was not found."
	   }
	}

	writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function stopProcesses() {
    writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    $amProcessCount = (Get-Process -Name "AutomationManager.*" | Measure-Object).count
    If ($amProcessCount -eq 0) {
        writeToLog V "No AutomationManager processes running."
    } Else {
        writeToLog W "Found $amProcessCount AutomationManager processes running, will now be stopped."

        try {
            Get-Process -Name "AutomationManager.*" | Stop-Process -Force -ErrorAction Stop
        } catch {
            $msg = $_.Exception.Message
            $line = $_.InvocationInfo.ScriptLineNumber
            writeToLog F "Error occurred when stopping AutomationManager process(es), due to:`r`n`t$msg"
            writeToLog V "This occurred on line number: $line"
            writeToLog F "Failing script."
            Exit 1001
        }
    }

    Start-Sleep 30

    $amProcessCount = (Get-Process -Name "AutomationManager.*" | Measure-Object).count

    If ($amProcessCount -eq 0) {
        writeToLog V "Confirmed no more AutomationManager processes are running."
    } Else {
        writeToLog F "Not all AutomationManager processes were stopped, remaining - $amProcessCount"
        writeToLog F "Failing script."
        Exit 1001
    }

	writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function cleanupAM() {
    writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    $amEnginePath = (Split-Path $localFolder) + "\AutomationManagerEngine\"

    If (!(Test-Path $amEnginePath)) {
        writeToLog V "The following folder does not exist:`r`n`t$amEnginePath"
    } Else {
        writeToLog I "The following folder does exist, so will clean up:`r`n`t$amEnginePath"

        try {
            Remove-Item $amEnginePath -Recurse -Force -ErrorAction Stop
        } catch {
            $msg = $_.Exception.Message
            $line = $_.InvocationInfo.ScriptLineNumber
            writeToLog F "Error occurred when deleting the AutomationManagerEngine folder contents, due to:`r`n`t$msg"
            writeToLog V "This occurred on line number: $line"
            writeToLog F "Failing script."
            Exit 1001
        }
    }

    $amConfigPath = "C:\ProgramData\N-Able Technologies\AutomationManager\upgrade.config"

    If (!(Test-Path $amConfigPath)) {
        writeToLog V "The following file does not exist:`r`n`t$amConfigPath"
    } Else {
        writeToLog I "The following file does exist, so will clean up:`r`n`t$amConfigPath"

        try {
            Remove-Item $amConfigPath -Recurse -Force -ErrorAction Stop
        } catch {
            $msg = $_.Exception.Message
            $line = $_.InvocationInfo.ScriptLineNumber
            writeToLog F "Error occurred when deleting the upgrade.config file, due to:`r`n`t$msg"
            writeToLog V "This occurred on line number: $line"
            writeToLog F "Failing script."
            Exit 1001
        }
    }

    $script:amAgentPath = (Split-Path $localFolder) + "\AutomationManagerAgent\"
    $dllsToKeep = "DotNetZip.dll", "Microsoft.PowerShell.Commands.Diagnostics.dll", "System.Management.Automation.dll", "System.ValueTuple.dll", "WPFToolkit.dll"

    writeToLog V "Cleaning up the Automation Manager Agent folder."

    try {
        Get-ChildItem $amAgentPath -Exclude $dllsToKeep | Remove-Item -Recurse -Force -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog F "Error occurred when cleaniing up the AutomationManagerAgent folder, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        writeToLog F "Failing script."
        Exit 1001
    }

    writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function downloadBinaries() {
    writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    $script:downloadPath = $logFolder + "\binaries.zip"
    $binariesUrl = "https://sis.n-able.com/Components/AutomationManagerEngine/2.60.0.3/AutomationManager-Binaries.zip"

    try {
        Invoke-WebRequest $binariesUrl -OutFile $downloadPath -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog F "Error occurred when downloading the AM binaries from sis.n-able.com, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        writeToLog F "Failing script."
        Exit 1001
    }

    writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function extractBinaries() {
    writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    writeToLog V "Copying downloaded binaries to the above path."

    try {
        Expand-Archive $downloadPath $amAgentPath -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog F "Error occurred when extracting the AM binaries, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
        writeToLog F "Failing script."
        Exit 1001
    }

    writeToLog V "Deleting the downloaded binaries.zip file."

    try {
        Remove-Item $downloadPath -Recurse -ErrorAction Stop
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog W "Error occurred when deleting the downloaded binaries.zip file, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
    }

    writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function populateAMReg() {
	writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

    $amAgentServicePath = $amAgentPath + "AutomationManager.AgentService.exe"

    try {
        $amAgentVersion = (Get-Item $amAgentServicePath -ErrorAction Stop).VersionInfo.FileVersion
    } catch {
        $msg = $_.Exception.Message
        $line = $_.InvocationInfo.ScriptLineNumber
        writeToLog W "Error occurred when obtaining the FileVersion from AutomationManager.AgentService.exe, due to:`r`n`t$msg"
        writeToLog V "This occurred on line number: $line"
    }

    If ($null -ne $amAgentVersion) {
        writeToLog I "New version of Automation Manager Agent: $amAgentVersion"

        try {
            Set-ItemProperty -Path $amRegLocation -Name "AutomationManagerVersion" -Value $amAgentVersion
        } catch {
            $msg = $_.Exception.Message
            $line = $_.InvocationInfo.ScriptLineNumber
            writeToLog W "Error occurred when updating the registry value for AutomationManagerVersion, due to:`r`n`t$msg"
            writeToLog V "This occurred on line number: $line"
        }
    }
    
	writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function startServices() {
	writeToLog V ("Started running {0} function." -f $MyInvocation.MyCommand)

	$array = @()
    If ($applianceType -eq "Sgent") {
        $array += "Windows Agent Maintenance Service"
        $array += "Windows Agent Service"
    } ElseIf ($applianceType -eq "Probe") {
        $array += "Windows Software Probe Maintenance Service"
        $array += "Windows Software Probe Service"
    }
	$array += "AutomationManagerAgent"
	
	foreach ($serviceName in $array) {

		If (Get-Service $serviceName -ErrorAction SilentlyContinue) {
			writeToLog I "Detected the $serviceName service on the device, will now start the service."

			try {
   				$script:startService = Start-Service -Name $serviceName -ErrorAction Stop -WarningAction SilentlyContinue
  			} catch {
				$msg = $_.Exception
				$line = $_.InvocationInfo.ScriptLineNumber
				writeToLog W "Failed to start $serviceName, due to:`r`n`t$($msg.Message)"
				writeToLog V "This occurred on line number: $line"
				writeToLog V "Status:`r`n`t$($msg.Status)`r`nResponse:`r`n`t$($msg.Response)`r`nInner Exception:`r`n`t$($msg.InnerException)`r`n`r`nHResult: $($msg.HResult)`r`n`r`nTargetSite and StackTrace:`r`n$($msg.TargetSite)`r`n$($msg.StackTrace)`r`n"
			}

			writeToLog I "$serviceName service is now started."

		} Else {
			writeToLog W "$serviceName service was not found."
	   }
	}

	writeToLog V ("Completed running {0} function." -f $MyInvocation.MyCommand)
}

function writeToLog($state, $message) {

    $script:timestamp = "[{0:dd/MM/yy} {0:HH:mm:ss}]" -f (Get-Date)

	switch -regex -Wildcard ($state) {
		"I" {
			$state = "INFO"
            $colour = "Cyan"
		}
		"E" {
			$state = "ERROR"
            $colour = "Red"
		}
		"W" {
			$state = "WARNING"
            $colour = "Yellow"
		}
		"F"  {
			$state = "FAILURE"
            $colour = "Red"
        }
        "C"  {
			$state = "COMPLETE"
            $colour = "Green"
        }
        "V"  {
            If ($verboseMode -eq $true) {
                $state = "VERBOSE"
                $colour = "Magenta"
            } Else {
                return
            }
		}
		""  {
			$state = "INFO"
		}
		Default {
			$state = "INFO"
		}
     }

    Write-Host "$($timeStamp) - [$state]: $message" -ForegroundColor $colour
    Write-Output "$($timeStamp) - [$state]: $message" | Out-file $logFilePath -Append -ErrorAction SilentlyContinue
}

function main() {
    setupLogging
    validateUserInput
    getAgentPath $applianceType
    stopProcesses
    stopServices
    stopProcesses
    cleanupAM
    downloadBinaries
    extractBinaries
    populateAMReg
    startServices
}
main